{:
PURPOSE AND IMPLEMENTATION:
  This unit defines a descendant of TGraphic, TDxf for loading and
  viewing .DXF files exported from AutoCAD.  This unit has limited support for
  .DXF files.  At this time it supports these entities:  lines, circles, arcs,
  lwpolylines, layers, and linetypes.  Support for linetypes could be more robust.
  Support for 3D is non-existent.  It also supports saving the DXF to a stream
  in a more compact/non-portable format useful for saving disk space.  Not only
  is this format more compact than a .DXF file but in loads in a fraction of the
  time.  This code is fairly well optimized for speed.

USAGE NOTES:
  This class supports printing via Nevrona's ReportPrinter Pro.  If you
  do not have this package installed you will want to disable printing
  support.  To do this comment out "RPDefine" and "RPBase" in the interface
  uses clause.  Also comment out TDxf.Print (both in the class definition
  and the implementation portion of this unit).

HISTORY:
  05/??/96 - Created by Colin Patrick Sarsfield (colin@csarsfield.com)
  10/17/96 - Print routine added by CPS
  10/23/96 - FillStrings routine added by CPS
  12/09/96 - LoadFromStreamCompact & SaveToStreamCompact added by CPS
  01/17/97 - Assign method added by CPS
  05/16/00 - Updated comments to new standard by CPS
  06/13/02 - Added support for lwpolylines and improved support for linetypes
             by CPS and Andrew Cray Vogan

COPYRIGHT:
  Released into the public domain.
}
unit Dxfgrph;

interface

uses
  Windows, Graphics, Classes
{$IFDEF DXFPRINT}
  ,RPDefine, RPBase
{$ENDIF}
  ;

type
  { this record is used to share stream data between functions }
  TIntDxfFile = packed record
    Stream: TStream;
    Buffer: PChar;
    BufferLen: Word;
    Position: Longint;
    LineNo: Longint;
    Size: Longint;
    Offset: Word;
    Eof: Boolean;
    Group: Integer;
    IVal: Integer;
    FVal: Single;
    SVal: string;
  end;
	{ TPoint didn't have a z }
  PXYZPoint = ^TXYZPoint;
  TXYZPoint = packed record
    x, y, z: Single;
  end;
	{ A drawing palette for the Dxf
  Actually, dxf's can have pens with colors from 0..255 so I guess this should
  probably be expanded.
  }
  TColorArray = packed array[0..15] of TColor;

  TIntDxfVertex = packed record
    P: TXYZPoint;
    StartWidth,
    EndWidth: Single;
  end;
  PIntDxfVertexArray = ^TIntDxfVertexArray;
  TIntDxfVertexArray = array[0..10000] of TIntDxfVertex;

  TDxfEntityKind = (dxfLine, dxfCircle, dxfArc, dxfLWPolyline);
  { TDxf Entity storage to internal use }
  TIntDxfEntityVariants = packed record
		case Kind: TDxfEntityKind of
      dxfCircle: (cCenter: TXYZPoint; cRadius: Single);
      dxfArc: (aCenter: TXYZPoint; aRadius: Single; BegAngle, EndAngle: Single);
      dxfLine: (Point1, Point2: TXYZPoint);
      dxfLWPolyline: (lpVertexCount: SmallInt; lpClosed: Boolean; lpVertices: PIntDxfVertexArray);
  end;
  TIntDxfEntity = packed record
    Layer: Smallint;
    V: TIntDxfEntityVariants;
    LineType: Smallint;
	end;
  { TDxf line type storage to internal use }
  TIntDxfLineType = packed record
    Name: ShortString;
    Style: TPenStyle;
  end;
  { TDxf layer storage to internal use }
  TIntDxfLayer = packed record
    Name: ShortString;
    Color: Smallint;
    LineType: Smallint;
  end;
  { pointer types for the above }
  PIntDxfEntity = ^TIntDxfEntity;
  PIntDxfLineType = ^TIntDxfLineType;
  PIntDxfLayer = ^TIntDxfLayer;

  { internal structure used by ZoomToExtents and ResetOrigin }
  PIntMinMaxPoints = ^TIntMinMaxPoints;
  TIntMinMaxPoints = record
    Min,
    Max: TXYZPoint;
  end;

  TPointEvent = procedure(var P: TXYZPoint; const Data: Pointer) of object;

  { Here it is! }
  TDxf = class(TGraphic)
  private
    FMarkOrigin: Boolean;
    FMonochromeDrawing: Boolean;

    procedure FreeEntity(PEntity: PIntDxfEntity);
    procedure GetGroup(var AFile: TIntDxfFile);
    function GetMappingDepth: Single;
    function GetMappingHeight: Single;
    function GetMappingWidth: Single;
    function ReadLine(var AFile: TIntDxfFile): string;
  protected
    FAngleBase: Single;
    FAngleDirection: Smallint;
    FDrawList: TList;
    FDrawingColors: TColorArray;
    FEmpty: Boolean;
    FHeight: Integer;
    FLineTypeList: TList;
		FLayerList: TList;
    FMaxDrawingExtent: TXYZPoint;
    FMaxMappingExtent: TXYZPoint;
    FMinDrawingExtent: TXYZPoint;
		FMinMappingExtent: TXYZPoint;
    FScaleFactor: Single;
    FWidth: Integer;
    procedure AdjustMinMaxForPoint(var P: TXYZPoint; const Data: Pointer);
    procedure CalcLineWithThicknessPolygon(const p1, p2: TXYZPoint;
      const Width1, Width2: Single; out SidePoints: array of TXYZPoint);
    procedure ClearLists;
    procedure Draw(ACanvas: TCanvas; const ARect: TRect); override;
    procedure EnumerateEntityPoints(const Entity: PIntDxfEntity; Data: Pointer;
      Callback: TPointEvent);
    function GetDrawList(Idx: Integer): PIntDxfEntity;
    function GetEmpty: Boolean; override;
    procedure GetExtents(out MinMax: TIntMinMaxPoints);
    function GetHeight: Integer; override;
    function GetLayer(Name: string): Integer;
    function GetLayerList(Idx: Integer): PIntDxfLayer;
    function GetLineType(const Name: string; const Default: Integer): Integer;
    function GetLineTypeList(Idx: Integer): PIntDxfLineType;
    function GetPenStyleByLayerAndLineType(Layer, LineType: SmallInt): TPenStyle;
    function GetWidth: Integer; override;
    procedure LWPolylineChangeVertexCount(const Entity: Pointer; const VCount: Integer);
    procedure MovePoint(var P: TXYZPoint; const Data: Pointer);
    procedure Scale;
    procedure ScalePoint(var P: TXYZPoint; const Data: Pointer);
    procedure SetHeight(Value: Integer); override;
    procedure SetScaleFactor(Value: Single);
    procedure SetMaxMappingExtent(Value: TXYZPoint);
    procedure SetMinMappingExtent(Value: TXYZPoint);
    procedure SetWidth(Value: Integer); override;

    property DrawList[Idx: Integer]: PIntDxfEntity read GetDrawList;
    property LineTypeList[Idx: Integer]: PIntDxfLineType read GetLineTypeList;
    property LayerList[Idx: Integer]: PIntDxfLayer read GetLayerList;
	public
    constructor Create; override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
		procedure DecEntitiesX(MaxX, XDec: Single);
    procedure DeleteEntitiesInRange(FromPoint, ToPoint: TXYZPoint);
    procedure FillStrings(Strings: TStrings);
    procedure IncEntitiesX(MinX, XInc: Single);
    procedure LoadFromClipboardFormat(AFormat: Word; AData: THandle;
      APalette: HPALETTE); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure LoadFromStreamCompact(Stream: TStream);
{$IFDEF DXFPRINT}
    procedure Print(Report: TBaseReport; Sideways: Boolean; X, Y, Scale: Single);
{$ENDIF}
    procedure ResetOrigin;
    procedure SaveToClipboardFormat(var AFormat: Word; var AData: THandle;
      var APalette: HPALETTE); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure SaveToStreamCompact(Stream: TStream);
    procedure ScaleBy(Factor: TXYZPoint);
    procedure TransformBy(Delta: TXYZPoint);
    procedure ZoomToExtents;

    property DrawingColors: TColorArray read FDrawingColors write FDrawingColors;
    property MappingDepth: Single read GetMappingDepth;
    property MappingHeight: Single read GetMappingHeight;
    property MappingWidth: Single read GetMappingWidth;
    property MarkOrigin: Boolean read FMarkOrigin write FMarkOrigin;
    property MaxDrawingExtent: TXYZPoint read FMaxDrawingExtent write FMaxDrawingExtent;
    property MaxMappingExtent: TXYZPoint read FMaxMappingExtent write FMaxMappingExtent;
    property MinDrawingExtent: TXYZPoint read FMinDrawingExtent write FMinDrawingExtent;
		property MinMappingExtent: TXYZPoint read FMinMappingExtent write FMinMappingExtent;
    property MonochromeDrawing: Boolean read FMonochromeDrawing write FMonochromeDrawing;
    property ScaleFactor: Single read FScaleFactor write SetScaleFactor;
  end;

{$IFDEF DXFPRINT}
procedure PrintDottedLine(Report: TBaseReport; x1, y1, x2, y2, DashLength: Double);
{$ENDIF}
function XYZPoint(x, y, z: Single): TXYZPoint;

implementation

uses
  SysUtils, StrLib, Math;

const
  DXFFileID = 293984383;
  DXFFileVer = 2;
  DashLength: Double = 0.03;
  LayerLineType = -1;

  { buffer size used by TDxf.LoadFromStream }
  BufferSize = 16 * 1024;

{ misc. functions }

{$IFDEF DXFPRINT}

// print the circle at (x, y) with radius r with dotted lines with each dot
// DashLength long.  Note that using the Ellipse function with the Pen.Style
// set to psDot does not do this as HP Deskjets and perhaps some other printers
// draw solid lines using this API.
procedure PrintDottedCircle(Report: TBaseReport; x, y, r, DashLength: Double);
var
  ang, ange, angi, circum: Double;
  i, NumDashes: Integer;
  Dash: Boolean;
begin
  with Report do begin
    // make sure radius is positive
    r := Abs(r);
    // calculate circumfrence
    circum := 2 * Pi * r;
    // start at angle 0
    ang := 0;
    // use an integral number of dashes / spaces (makes the code simpler
    // the circle probably looks better this way too)
    NumDashes := Round(0.5 * circum / DashLength) * 2;
    if NumDashes = 0 then Exit;
    // calculate length of each dash
    angi := 2 * Pi / NumDashes;

    Dash := True;
    // draw each dash
    for i := 1 to NumDashes do begin
      ange := ang + angi;
      if Dash then
        Arc(x - r, y + r, x + r, y - r, x + r * cos(ang), y - r * sin(ang),
             x + r * cos(ange), y - r * sin(ange));
      ang := ange;
      Dash := not Dash;
    end;
  end;
end;

// draw a dotted arc from angle bang to angle eang on the circle at (x, y) with
// radius r.  Make each dash DashLength long.  Note that using the Arc function
// with the Pen.Style set to psDot does not do this as HP Deskjets and perhaps
// some other printers draw solid lines using this API.
procedure PrintDottedCirclularArc(Report: TBaseReport; x, y, r, bang, eang, DashLength: Double);
var
  ang, ange, angi, circum: Double;
  i, NumDashes: Integer;
  Dash: Boolean;
begin
  with Report do begin
    // no negative radiuses
    r := Abs(r);
    // this is probably overkill, but it works
    bang := 2 * Pi * Frac(bang / (2 * Pi));
    eang := 2 * Pi * Frac(eang / (2 * Pi));
    if bang > eang then eang := eang + 2 * Pi;

    // use the same calcs you would for a circle but ratio them to the angle
    // being shown
    circum := 2 * Pi * r;
    ang := bang;
    NumDashes := Round((eang - bang) / (2 * Pi) * circum / DashLength);
    if NumDashes = 0 then Exit;
    angi := (eang - bang) / NumDashes;

    Dash := True;
    for i := 1 to NumDashes do begin
      ange := ang + angi;
      if Dash then
        Arc(x - r, y + r, x + r, y - r, x + r * cos(ang), y - r * sin(ang),
            x + r * cos(ange), y - r * sin(ange));
      ang := ange;
      Dash := not Dash;
    end;
  end;
end;

// print the line from (x1, y1) to (x2, y2)with dotted lines with each dot
// DashLength long.  Note that using the MoveTo/LineTo functions with the Pen.Style
// set to psDot does not do this as HP Deskjets and perhaps some other printers
// draw solid lines using this API.
procedure PrintDottedLine(Report: TBaseReport; x1, y1, x2, y2, DashLength: Double);
var
  xi, yi, linelen: Double;
  i, DashCntM1: Integer;
  Dash: Boolean;
begin
  with Report do begin
    MoveTo(x1, y1);
    linelen := Sqrt(Sqr(x2 - x1) + Sqr(y2 - y1)); // line length
    if linelen = 0 then Exit;
    xi := DashLength * (x2 - x1) / linelen; // dash inc in x
    yi := DashLength * (y2 - y1) / linelen; // dash inc in y
    Dash := True;
    DashCntM1 := Trunc(linelen / DashLength);
    for i := 1 to DashCntM1 do begin
      x1 := x1 + xi;
      y1 := y1 + yi;
      if Dash then
        LineTo(x1, y1)
      else
        MoveTo(x1, y1);
      Dash := not Dash;
    end;
    // last dash
    if Dash then
      LineTo(x2, y2)
    else
      MoveTo(x2, y2);
  end;
end;

{$ENDIF}


{: Given the vector defined from (x1, y1) to (x2, y2), this procedure
returns in ResultX and ResultY the endpoint of a vector originating from (x1, y2),
having a length of Dist and offset from the first vector by angle Theta.}
procedure CalcPointOutsideVector(const x1, y1, x2, y2, Theta, Dist: Single;
  out ResultX, ResultY: Single);
var
  Angle: Single;
begin
  if IsZero(x2 - x1) then begin
    if y2 > y1 then
      Angle := PI / 2
    else
      Angle := -PI / 2;
  end else begin
    Angle := ArcTan((y2 - y1) / (x2 - x1));
    if x2 < x1 then
      Angle := Angle + PI;
  end;

  ResultX := x1 + Dist * Cos(Theta + Angle);
  ResultY := y1 + Dist * Sin(Theta + Angle);
end;

{ just an easy way to fill a TXYZPoint structure }

function XYZPoint(x, y, z: Single): TXYZPoint;
begin
  Result.x := x;
  Result.y := y;
	Result.z := z;
end;

{ TDxf }

constructor TDxf.Create;
const
	DefaultColors: TColorArray = (
		clWindowText, clWindowText, clWindowText, clHighlight, clHighlight, clWindowText, clWindowText, clWindowText,
		clWindowText, clWindowText, clWindowText, clWindowText, clWindowText, clWindowText, clWindowText, clWindowText);
begin
  inherited Create;
  FDrawList := TList.Create;
  FLineTypeList := TList.Create;
  FLayerList := TList.Create;
  FDrawingColors := DefaultColors;
  FEmpty := True;
  FScaleFactor := 1;
end;

destructor TDxf.Destroy;
begin
  ClearLists;
  FDrawList.Free;
  FLineTypeList.Free;
  FLayerList.Free;
  inherited Destroy;
end;

procedure TDxf.AdjustMinMaxForPoint(var P: TXYZPoint; const Data: Pointer);
var
  MM: PIntMinMaxPoints;
begin
  MM := PIntMinMaxPoints(Data);

  MM.Min.x := Min(MM.Min.x, P.x);
  MM.Min.y := Min(MM.Min.y, P.y);
  MM.Min.z := Min(MM.Min.z, P.z);

  MM.Max.x := Max(MM.Max.x, P.x);
  MM.Max.y := Max(MM.Max.y, P.y);
  MM.Max.z := Max(MM.Max.z, P.z);
end;

procedure TDxf.Assign(Source: TPersistent);
var
	Dxf: TDxf;
  NewEntity: PIntDxfEntity;
  NewLineType: PIntDxfLineType;
  NewLayer: PIntDxfLayer;
  i: Integer;
begin
	if Source is TDxf then
  	Dxf := TDxf(Source)
  else
  	raise Exception.Create('TDxf.Assign can only assign from other TDxf''s');
  ClearLists;
  if Dxf.Empty then Exit;
  FAngleBase := Dxf.FAngleBase;
  FAngleDirection := Dxf.FAngleDirection;
  for i := 0 to Dxf.FDrawList.Count - 1 do begin
    New(NewEntity);
    NewEntity^ := Dxf.DrawList[i]^;
    if NewEntity^.V.Kind = dxfLWPolyline then begin
      NewEntity^.V.lpVertices := nil;
      LWPolylineChangeVertexCount(NewEntity, NewEntity^.V.lpVertexCount);
      Move(Dxf.DrawList[i]^.V.lpVertices^, NewEntity^.V.lpVertices^,
          NewEntity^.V.lpVertexCount * SizeOf(TIntDxfVertex));
    end;
    FDrawList.Add(NewEntity);
  end;
  for i := 0 to Dxf.FLineTypeList.Count - 1 do begin
    New(NewLineType);
    NewLineType^ := PIntDxfLineType(Dxf.FLineTypeList[i])^;
    FLineTypeList.Add(NewLineType);
  end;
  for i := 0 to Dxf.FLayerList.Count - 1 do begin
    New(NewLayer);
    NewLayer^ := PIntDxfLayer(Dxf.FLayerList[i])^;
    FLayerList.Add(NewLayer);
  end;
  FMaxDrawingExtent := Dxf.FMaxDrawingExtent;
  FMinDrawingExtent := Dxf.FMinDrawingExtent;
  FMaxMappingExtent := Dxf.FMaxMappingExtent;
  FMinMappingExtent := Dxf.FMinMappingExtent;
  FScaleFactor := Dxf.FScaleFactor;

  FEmpty := False;
  Scale; { calculate height and width, kept for compatibility with TGraphic }
  Changed(Self);
end;

procedure TDxf.CalcLineWithThicknessPolygon(const p1, p2: TXYZPoint;
  const Width1, Width2: Single; out SidePoints: array of TXYZPoint);
begin
  if (Low(SidePoints) <> 0) or (High(SidePoints) <> 3) then
    raise Exception.Create('Invalid point array in CalcLineWithThicknessPolygon');

  CalcPointOutsideVector(p1.x, p1.y, p2.x, p2.y, -PI / 2, Width1 / 2,
      SidePoints[0].x, SidePoints[0].y);
  CalcPointOutsideVector(p1.x, p1.y, p2.x, p2.y, PI / 2, Width1 / 2,
      SidePoints[1].x, SidePoints[1].y);

  CalcPointOutsideVector(p2.x, p2.y, p1.x, p1.y, -PI / 2, Width2 / 2,
      SidePoints[2].x, SidePoints[2].y);
  CalcPointOutsideVector(p2.x, p2.y, p1.x, p1.y, PI / 2, Width2 / 2,
      SidePoints[3].x, SidePoints[3].y);
end;

{ the method used to clear/empty a drawing }
procedure TDxf.ClearLists;
var
  i: Integer;
begin
  for i := 0 to FDrawList.Count - 1 do
    FreeEntity(DrawList[i]);
  FDrawList.Clear;
  for i := 0 to FLineTypeList.Count - 1 do
    Dispose(LineTypeList[i]);
  FLineTypeList.Clear;
  for i := 0 to FLayerList.Count - 1 do
    Dispose(LayerList[i]);
  FLayerList.Clear;
  FEmpty := True;
end;

{used to "move" entities x coords below "MaxX" by "XDec"}
procedure TDxf.DecEntitiesX(MaxX, XDec: Single);
var
  i, j: Integer;

  procedure MoveXCoord(var x: Single);
  begin
    if x < MaxX then begin
      x := x - XDec;
      if x < FMinMappingExtent.x then
        FMinMappingExtent.x := x;
    end;
  end;
begin
  FMinMappingExtent.x := FMaxMappingExtent.x;
  for i := 0 to FDrawList.Count - 1 do begin
    with DrawList[i]^ do begin
      case V.Kind of
      dxfCircle: begin
          if V.cCenter.x < MaxX then
            V.cCenter.x := V.cCenter.x - XDec;
        end;
      dxfArc: begin
          if V.aCenter.x < MaxX then
            V.aCenter.x := V.aCenter.x - XDec;
        end;
      dxfLine: begin
          MoveXCoord(V.Point1.x);
          MoveXCoord(V.Point2.x);
        end;
      dxfLWPolyline:
        for j := 0 to Pred(V.lpVertexCount) do
          MoveXCoord(V.lpVertices^[j].P.x);
      end;
    end;
  end;
  FMinDrawingExtent.x := FMinMappingExtent.x;
end;

{ used by TCanvas.Draw for drawing }
procedure TDxf.Draw(ACanvas: TCanvas; const ARect: TRect);
var
	mx, bx, my, by: Single;
	BAngle, EAngle, SWidth: Single;
	Layer, i, j: Integer;
  PenLocation: TXYZPoint;

  procedure IntMoveTo(const p: TXYZPoint; const StartWidth: Single);
  begin
    SWidth := StartWidth;
    PenLocation := p;
    ACanvas.MoveTo(
        Round(p.x * mx + bx),
        Round(p.y * my + by));
  end;

  procedure IntLineTo(const p: TXYZPoint; const EndWidth, NextStartWidth: Single);
  var
  	ix, iy: Integer;

    procedure DrawLineWithThickness(const p1, p2: TXYZPoint;
      const Width1, Width2: Single);
    var
      m: Integer;
      SidePoints: array[0..3] of TXYZPoint;
      CanvasSidePoints: array[0..3] of TPoint;
    begin
      CalcLineWithThicknessPolygon(p1, p2, Width1, Width2, SidePoints);

      for m := Low(SidePoints) to High(SidePoints) do
        CanvasSidePoints[m] := Point(Round(SidePoints[m].x * mx + bx),
            Round(SidePoints[m].y * my + by));

      ACanvas.Brush.Style := bsSolid;
      ACanvas.Brush.Color := ACanvas.Pen.Color;

      ACanvas.Polygon(CanvasSidePoints);

      ACanvas.Brush.Style := bsClear;
    end;
  begin
    ix := Round(p.x * mx + bx);
    iy := Round(p.y * my + by);
    if (SWidth = 0) and (EndWidth = 0) then begin
      ACanvas.LineTo(ix, iy);
      ACanvas.Pixels[ix, iy] := ACanvas.Pen.Color;
    end else begin
      DrawLineWithThickness(PenLocation, p, SWidth, EndWidth);
    end;
    IntMoveTo(p, NextStartWidth);
  end;
begin
	if Empty then Exit;
	mx := (ARect.Right - ARect.Left - 1) / (FMaxMappingExtent.x - FMinMappingExtent.x);
	my := (ARect.Top - ARect.Bottom + 1) / (FMaxMappingExtent.y - FMinMappingExtent.y);
	bx := ARect.Left - FMinMappingExtent.x * mx;
	by := ARect.Bottom - FMinMappingExtent.y * my - 1;
	ACanvas.Brush.Style := bsClear;
	Layer := -1;
  if MarkOrigin then begin
    ACanvas.Pen.Style := psSolid;
    if not MonochromeDrawing then ACanvas.Pen.Color := clRed;
    ACanvas.MoveTo(Round(bx) - 200, Round(by) - 200);
    ACanvas.LineTo(Round(bx) + 200, Round(by) + 200);
    ACanvas.MoveTo(Round(bx) + 200, Round(by) - 200);
    ACanvas.LineTo(Round(bx) - 200, Round(by) + 200);
  end;
	for i := 0 to FDrawList.Count - 1 do begin
		if Layer <> DrawList[i]^.Layer then begin
			Layer := DrawList[i]^.Layer;
			if not MonochromeDrawing then
        ACanvas.Pen.Color := FDrawingColors[LayerList[Layer]^.Color mod 16];
		end;
  	ACanvas.Pen.Style := GetPenStyleByLayerAndLineType(Layer, DrawList[i]^.LineType);
		with DrawList[i]^ do
      case V.Kind of
      dxfCircle: begin
          ACanvas.Ellipse(
              Round((V.cCenter.x - V.cRadius) * mx + bx),
              Round((V.cCenter.y - V.cRadius) * my + by),
              Round((V.cCenter.x + V.cRadius) * mx + bx),
              Round((V.cCenter.y + V.cRadius) * my + by));
        end;
			dxfArc: begin
          if FAngleDirection = 1 then begin
            BAngle := -V.BegAngle;
            EAngle := -V.EndAngle;
          end
          else begin
            BAngle := V.BegAngle;
            EAngle := V.EndAngle;
          end;
          BAngle := (BAngle + FAngleBase) / 180 * Pi;
          EAngle := (EAngle + FAngleBase) / 180 * Pi;
          ACanvas.Arc(
              Round((V.aCenter.x - V.aRadius) * mx + bx),
              Round((V.aCenter.y - V.aRadius) * my + by),
              Round((V.aCenter.x + V.aRadius) * mx + bx),
              Round((V.aCenter.y + V.aRadius) * my + by),
              Round((V.aCenter.x + V.aRadius * cos(BAngle)) * mx + bx),
              Round((V.aCenter.y + V.aRadius * sin(BAngle)) * my + by),
              Round((V.aCenter.x + V.aRadius * cos(EAngle)) * mx + bx),
              Round((V.aCenter.y + V.aRadius * sin(EAngle)) * my + by)
              );
			end;
			dxfLine: begin
          IntMoveTo(V.Point1, 0);
          IntLineTo(V.Point2, 0, 0);
        end;
      dxfLWPolyline:
        if V.lpVertexCount > 0 then begin
          IntMoveTo(V.lpVertices^[0].P, V.lpVertices^[0].StartWidth);
          for j := 1 to Pred(V.lpVertexCount) do
            IntLineTo(V.lpVertices^[j].P, V.lpVertices^[j].EndWidth,
                V.lpVertices^[j].StartWidth);
          if V.lpClosed then
            IntLineTo(V.lpVertices^[0].P, V.lpVertices^[0].EndWidth, 0);
        end;
      end;
	end;
end;

procedure TDxf.EnumerateEntityPoints(const Entity: PIntDxfEntity;
  Data: Pointer; Callback: TPointEvent);
var
  i: Integer;

  procedure DoCallback(var p: TXYZPoint);
  begin
    Callback(p, Data);
  end;
begin
  with Entity^ do
    case V.Kind of
    dxfLine: begin
        DoCallback(V.Point1);
        DoCallback(V.Point2);
      end;
    dxfCircle:
      DoCallback(V.cCenter);
    dxfArc:
      DoCallback(V.aCenter);
    dxfLWPolyline:
      for i := 0 to Pred(V.lpVertexCount) do
        DoCallback(V.lpVertices^[i].P);
    end;
end;

{ for debugging/research uses (to see what TDxf has read in) }
procedure TDxf.FillStrings(Strings: TStrings);
const
  CloseStr: array[Boolean] of String = ('Open', 'Closed');
var
  Layer, i, j: Integer;
  WorkStr: String;
begin
  if Empty then Exit;
  Layer := -1;
  Strings.Clear;
  for i := 0 to FDrawList.Count - 1 do begin
		if Layer <> DrawList[i]^.Layer then begin
      Layer := DrawList[i]^.Layer;
      Strings.Add('Layer: ' + TIntDxfLayer(FLayerList[Layer]^).Name);
    end;
    with DrawList[i]^ do
      case V.Kind of
			dxfCircle:
				Strings.Add(Format('Circle: (%g, %g) - %g', [V.cCenter.x, V.cCenter.y, V.cRadius]));
			dxfArc:
				Strings.Add(Format('Arc: (%g, %g) - %g : %g - %g', [V.aCenter.x, V.aCenter.y, V.aRadius, V.BegAngle, V.EndAngle]));
			dxfLine:
				Strings.Add(Format('Line: (%g, %g) - (%g, %g)', [V.Point1.x, V.Point1.y, V.Point2.x, V.Point2.y]));
      dxfLWPolyline: begin
          WorkStr := 'LWPolyline: ';
          for j := 0 to Pred(V.lpVertexCount) do begin
            if j <> 0 then
              WorkStr := WorkStr + ' - ';
            WorkStr := WorkStr + Format('(%g, %g)', [V.lpVertices^[j].P.x, V.lpVertices^[j].P.y]);
          end;
          WorkStr := WorkStr + ' ' + CloseStr[V.lpClosed];
          Strings.Add(WorkStr);
        end;
      end;
	end;
end;

function TDxf.GetDrawList(Idx: Integer): PIntDxfEntity;
begin
  Result := PIntDxfEntity(FDrawList[Idx]);
end;

function TDxf.GetEmpty: Boolean;
begin
	Result := FEmpty;
end;

procedure TDxf.GetExtents(out MinMax: TIntMinMaxPoints);
var
  i: Integer;
begin
  with MinMax do begin
    Min.x := MaxSingle;
    Min.y := MaxSingle;
    Min.z := MaxSingle;

    Max.x := -MaxSingle;
    Max.y := -MaxSingle;
    Max.z := -MaxSingle;
  end;

  for i := 0 to Pred(FDrawList.Count) do
    EnumerateEntityPoints(DrawList[i], @MinMax, AdjustMinMaxForPoint);
end;

function TDxf.GetHeight: Integer;
begin
  Result := FHeight;
end;

{ Input procedure used by LoadFromStream - see AutoCAD's dxf file format
documentation for a description of groups }

procedure TDxf.GetGroup(var AFile: TIntDxfFile);
var
  TStr: string;
  TryCount, XCode: Integer;
begin
  TryCount := 0;
  repeat
    TStr := ReadLine(AFile);
    Val(TStr, AFile.Group, XCode);
		Inc(TryCount);
    if TryCount > 3 then
      raise Exception.Create('Invalid DXF file');
  until XCode = 0;
  if ((AFile.Group >= 0) and (AFile.Group <= 9))
    or ((AFile.Group >= 999) and (AFile.Group <= 1009)) then begin
    try
      AFile.SVal := ReadLine(AFile);
    except
      AFile.SVal := '';
    end;
  end
  else if ((AFile.Group >= 10) and (AFile.Group <= 59))
    or ((AFile.Group >= 140) and (AFile.Group <= 147))
    or ((AFile.Group >= 210) and (AFile.Group <= 239))
    or ((AFile.Group >= 1010) and (AFile.Group <= 1059)) then begin
    try
      AFile.FVal := IntlStrToFloat(ReadLine(AFile));
    except
      AFile.FVal := 0;
    end;
  end
  else if ((AFile.Group >= 60) and (AFile.Group <= 79))
    or ((AFile.Group >= 90) and (AFile.Group <= 99))
    or ((AFile.Group >= 170) and (AFile.Group <= 175))
    or ((AFile.Group >= 1060) and (AFile.Group <= 1075)) then begin
    try
      AFile.IVal := StrToInt(ReadLine(AFile));
    except
      AFile.IVal := 0;
    end;
  end
  else
    ReadLine(AFile);
end;

{ Layer lookup by name }

function TDxf.GetLayer(Name: string): Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to FLayerList.Count - 1 do
    if TIntDxfLayer(FLayerList[i]^).Name = Name then begin
      Result := i;
      Break;
    end;
end;

function TDxf.GetLayerList(Idx: Integer): PIntDxfLayer;
begin
  Result := PIntDxfLayer(FLayerList[Idx]);
end;

{ Line type lookup by name }

function TDxf.GetLineType(const Name: string; const Default: Integer): Integer;
var
  i: Integer;
begin
  Result := Default;
  for i := 0 to FLineTypeList.Count - 1 do
    if TIntDxfLineType(FLineTypeList[i]^).Name = Name then begin
      Result := i;
      Break;
    end;
end;

function TDxf.GetLineTypeList(Idx: Integer): PIntDxfLineType;
begin
  Result := PIntDxfLineType(FLineTypeList[Idx]);
end;

function TDxf.GetPenStyleByLayerAndLineType(Layer, LineType: SmallInt): TPenStyle;
begin
  if LineType = LayerLineType then
    LineType := LayerList[Layer]^.LineType;
  Result := LineTypeList[LineType]^.Style;
end;

function TDxf.GetWidth: Integer;
begin
  Result := FWidth;
end;

{used to "move" entities x coords above "MinX" by "XInc"}
procedure TDxf.IncEntitiesX(MinX, XInc: Single);
var
  i, j: Integer;

  procedure MoveXCoord(var x: Single);
  begin
    if x > MinX then
      x := x + XInc;
  end;
begin
  for i := 0 to FDrawList.Count - 1 do begin
    with DrawList[i]^ do
      case V.Kind of
      dxfCircle: MoveXCoord(V.cCenter.x);
      dxfArc: MoveXCoord(V.aCenter.x);
      dxfLine: begin
          MoveXCoord(V.Point1.x);
          MoveXCoord(V.Point2.x);
        end;
      dxfLWPolyline:
        for j := 0 to Pred(V.lpVertexCount) do
          MoveXCoord(V.lpVertices^[j].P.x);
      end;
  end;
end;

procedure TDxf.LoadFromClipboardFormat(AFormat: Word; AData: THandle;
      APalette: HPALETTE);
begin
	raise Exception.Create('TDxf does not support loading from clipboard format');
end;

{ This is where the DXF is loaded }

procedure TDxf.LoadFromStream(Stream: TStream);
var
  VertexIdx: Integer;
  AFile: TIntDxfFile;
  Section, Entity, TableType: string;
  CurEntity: PIntDxfEntity;
  CurLineType: PIntDxfLineType;
  CurLayer: PIntDxfLayer;
  IsDXF: Boolean;
  UCSOrg: TXYZPoint;
begin
  // initialized to silence compiler warning
  VertexIdx := -1000;

  { set stream }
  AFile.Stream := Stream;
  { move to beginning of stream }
  Stream.Seek(0, 0);
  AFile.LineNo := 1;
  AFile.Eof := False;
  AFile.Offset := 64000;
  AFile.BufferLen := BufferSize;
  AFile.Size := Stream.Size;
  AFile.Position := 0;
  GetMem(AFile.Buffer, BufferSize);
  try
    { initialize variables to default values }
    CurEntity := nil;
    CurLineType := nil;
    CurLayer := nil;
    Section := '';
    Entity := '';
    TableType := '';
    FAngleBase := 0;
    FAngleDirection := 0;
    UCSOrg := XYZPoint(0, 0, 0);
    FMinDrawingExtent := XYZPoint(0, 0, 0);
    FMaxDrawingExtent := XYZPoint(1, 1, 1);
    { clear lists of drawing objects }
    ClearLists;
    IsDxf := False;
    { parse Dxf file until end of file }
    while not AFile.Eof do begin
      { get "group" consisting of two lines of data:
       group #
       group data (string, integer, or floating-point based on group #)
       }
      GetGroup(AFile);
      if not IsDxf then begin
        if (AFile.Group = 0) and (AFile.SVal = 'SECTION') then
          IsDxf := True
        else
          raise Exception.Create('Invalid DXF file');
      end;
      with AFile do
        case Group of
          0, 9:
            { group 0 indicates the current part of the file }
            { group 9 indicates a variable in the header section }
            { they are handled similarly here to save on stack hungry string variables } begin
              { process last part }
              { the else if structure is used because the below can't occur
               concurrently }
              if CurEntity <> nil then begin
                FDrawList.Add(CurEntity);
                CurEntity := nil;
              end
              else if CurLineType <> nil then begin
                FLineTypeList.Add(CurLineType);
                CurLineType := nil;
              end
              else if CurLayer <> nil then begin
                FLayerList.Add(CurLayer);
                CurLayer := nil;
              end;
              {get new entity}
              Entity := SVal;
              {special group 0 processing}
              if Group = 0 then begin
                { process end of file }
                if Entity = 'EOF' then
                  { set eof flag }
                  AFile.Eof := True
                    { process end of section }
                else if Entity = 'ENDSEC' then
                  { clear section }
                  Section := ''
                    { process end of table type }
                else if Entity = 'ENDTAB' then
                  { clear table type }
                  TableType := '';
              end;
              { process new part }
              {create new structure to store data in and
              set defaults, just in case dxf provides incomplete data}
              if Section = 'ENTITIES' then begin
                CurEntity := nil;
                if Entity = 'CIRCLE' then begin
                  New(CurEntity);
                  CurEntity^.V.Kind := dxfCircle;
                  CurEntity^.V.cCenter := XYZPoint(0, 0, 0);
                  CurEntity^.V.cRadius := 0;
                end
                else if Entity = 'ARC' then begin
                  New(CurEntity);
                  CurEntity^.V.Kind := dxfArc;
                  CurEntity^.V.aCenter := XYZPoint(0, 0, 0);
                  CurEntity^.V.aRadius := 0;
                end
                else if Entity = 'LINE' then begin
                  New(CurEntity);
                  CurEntity^.V.Kind := dxfLine;
                  CurEntity^.V.Point1 := XYZPoint(0, 0, 0);
                  CurEntity^.V.Point2 := XYZPoint(0, 0, 0);
                end
                else if Entity = 'LWPOLYLINE' then begin
                  New(CurEntity);
                  CurEntity^.V.Kind := dxfLWPolyline;
                  CurEntity^.V.lpVertexCount := 0;
                  CurEntity^.V.lpClosed := False;
                  CurEntity^.V.lpVertices := nil;
                  VertexIdx := -1;
                end;
                if CurEntity <> nil then begin
                  CurEntity^.Layer := 0;
                  CurEntity^.LineType := LayerLineType;
                end;
              end
              else if Section = 'TABLES' then begin
                if Entity = 'LTYPE' then begin
                  New(CurLineType);
                  CurLineType^.Name := '';
                  CurLineType^.Style := psSolid;
                end
                else if Entity = 'LAYER' then begin
                  New(CurLayer);
                  CurLayer^.Name := '';
                  CurLayer^.Color := clBlack;
                  CurLayer^.LineType := 0;
                end;
              end;
            end; {groups 0 and 9}
          2: begin
              if Entity = 'SECTION' then
                Section := SVal
              else if Entity = 'TABLE' then
                TableType := SVal
              else if (TableType = 'LTYPE') and (Entity = 'LTYPE') then
                CurLineType^.Name := SVal
              else if (TableType = 'LAYER') and (Entity = 'LAYER') then
                CurLayer^.Name := SVal;
            end; {group 2}
          6: begin
              if (TableType = 'LAYER') and (Entity = 'LAYER') then
                CurLayer^.LineType := GetLineType(SVal, 0)
              else if (Section = 'ENTITIES') and (CurEntity <> nil) then
                CurEntity^.LineType := GetLineType(SVal, LayerLineType);
            end; {group 6}
          8: begin
              if (Section = 'ENTITIES') and ((Entity = 'LINE') or
                (Entity = 'ARC') or (Entity = 'CIRCLE')) then
                CurEntity^.Layer := GetLayer(SVal);
            end; {group 8}
          10: begin
              if Section = 'ENTITIES' then begin
                if Entity = 'LINE' then
                  CurEntity^.V.Point1.x := FVal - UCSOrg.x
                else if Entity = 'ARC' then
                  CurEntity^.V.aCenter.x := FVal - UCSOrg.x
                else if Entity = 'CIRCLE' then
                  CurEntity^.V.cCenter.x := FVal - UCSOrg.x
                else if Entity = 'LWPOLYLINE' then begin
                  Inc(VertexIdx);
                  if (VertexIdx >= 0) and (VertexIdx < CurEntity^.V.lpVertexCount) then
                    CurEntity^.V.lpVertices^[VertexIdx].P.x := FVal - UCSOrg.x;
                end;
              end
              else if (Section = 'HEADER') and (Entity = '$EXTMAX') then
                FMaxDrawingExtent.x := FVal
              else if (Section = 'HEADER') and (Entity = '$EXTMIN') then
                FMinDrawingExtent.x := FVal
              else if (Section = 'HEADER') and (Entity = '$UCSORG') then
                UCSOrg.x := FVal;
            end; {group 10}
          11: begin
              if (Section = 'ENTITIES') and (Entity = 'LINE') then
                CurEntity^.V.Point2.x := FVal - UCSOrg.x;
            end; {group 11}
          20: begin
              if Section = 'ENTITIES' then begin
                if Entity = 'LINE' then
                  CurEntity^.V.Point1.y := FVal - UCSOrg.y
                else if Entity = 'ARC' then
                  CurEntity^.V.aCenter.y := FVal - UCSOrg.y
                else if Entity = 'CIRCLE' then
                  CurEntity^.V.cCenter.y := FVal - UCSOrg.y
                else if Entity = 'LWPOLYLINE' then
                  if (VertexIdx >= 0) and (VertexIdx < CurEntity^.V.lpVertexCount) then
                    CurEntity^.V.lpVertices^[VertexIdx].P.y := FVal - UCSOrg.y;
              end
              else if (Section = 'HEADER') and (Entity = '$EXTMAX') then
                FMaxDrawingExtent.y := FVal
              else if (Section = 'HEADER') and (Entity = '$EXTMIN') then
                FMinDrawingExtent.y := FVal
              else if (Section = 'HEADER') and (Entity = '$UCSORG') then
                UCSOrg.y := FVal;
            end; {group 20}
          21: begin
              if (Section = 'ENTITIES') and (Entity = 'LINE') then
                CurEntity^.V.Point2.y := FVal - UCSOrg.y;
            end; {group 21}
          30: begin
              if Section = 'ENTITIES' then begin
                if Entity = 'LINE' then
                  CurEntity^.V.Point1.z := FVal - UCSOrg.z
                else if Entity = 'ARC' then
                  CurEntity^.V.aCenter.z := FVal - UCSOrg.z
                else if Entity = 'CIRCLE' then
                  CurEntity^.V.cCenter.z := FVal - UCSOrg.z
                else if Entity = 'LWPOLYLINE' then
                  if (VertexIdx >= 0) and (VertexIdx < CurEntity^.V.lpVertexCount) then
                    CurEntity^.V.lpVertices^[VertexIdx].P.z := FVal - UCSOrg.z;
              end
              else if (Section = 'HEADER') and (Entity = '$EXTMAX') then
                FMaxDrawingExtent.z := FVal
              else if (Section = 'HEADER') and (Entity = '$EXTMIN') then
                FMinDrawingExtent.z := FVal
              else if (Section = 'HEADER') and (Entity = '$UCSORG') then
                UCSOrg.z := FVal;
            end; {group 30}
          31: begin
              if (Section = 'ENTITIES') and (Entity = 'LINE') then
                CurEntity^.V.Point2.z := FVal - UCSOrg.z;
            end; {group 31}
          40: begin
              if Section = 'ENTITIES' then begin
                if Entity = 'ARC' then
                  CurEntity^.V.aRadius := FVal
                else if Entity = 'CIRCLE' then
                  CurEntity^.V.cRadius := FVal
                else if Entity = 'LWPOLYLINE' then
                  if (VertexIdx >= 0) and (VertexIdx < CurEntity^.V.lpVertexCount) then
                    CurEntity^.V.lpVertices^[VertexIdx].StartWidth := FVal;
              end;
            end; {group 40}
          41: begin
              if Section = 'ENTITIES' then begin
                if Entity = 'LWPOLYLINE' then
                  if (VertexIdx >= 0) and (VertexIdx < CurEntity^.V.lpVertexCount) then
                    CurEntity^.V.lpVertices^[VertexIdx].EndWidth := FVal;
              end;
            end; {group 41}
          50: begin
              if (Section = 'ENTITIES') and (Entity = 'ARC') then
                CurEntity^.V.BegAngle := FVal
              else if (Section = 'HEADER') and (Entity = '$ANGBASE') then
                FAngleBase := FVal;
            end; {group 50}
          51: begin
              if (Section = 'ENTITIES') and (Entity = 'ARC') then
                CurEntity^.V.EndAngle := FVal;
            end; {group 51}
          62: begin
              if (TableType = 'LAYER') and (Entity = 'LAYER') then
                CurLayer^.Color := IVal;
            end; {group 62}
          70: begin
              if (Section = 'ENTITIES') and (Entity = 'LWPOLYLINE') then
                CurEntity^.V.lpClosed := Boolean(IVal and 1)
              else if (Section = 'HEADER') and (Entity = '$ANGBASE') then
                FAngleDirection := IVal;
            end; {group 70}
          73: begin
              if (TableType = 'LTYPE') and (Entity = 'LTYPE') then begin
                if IVal = 0 then
                  CurLineType^.Style := psSolid
                else
                  CurLineType^.Style := psDot;
              end;
            end; {group 73}
          90: begin
              if (Section = 'ENTITIES') and (Entity = 'LWPOLYLINE') then
                LWPolylineChangeVertexCount(CurEntity, IVal);
            end; {group 90}
        end; {case}
    end;
    FMinDrawingExtent.x := FMinDrawingExtent.x - UCSOrg.x;
    FMinDrawingExtent.y := FMinDrawingExtent.y - UCSOrg.y;
    FMinDrawingExtent.z := FMinDrawingExtent.z - UCSOrg.z;
    FMaxDrawingExtent.x := FMaxDrawingExtent.x - UCSOrg.x;
    FMaxDrawingExtent.y := FMaxDrawingExtent.y - UCSOrg.y;
    FMaxDrawingExtent.z := FMaxDrawingExtent.z - UCSOrg.z;
    { set mapping extent to drawing extent for default }
    FMinMappingExtent := FMinDrawingExtent;
    FMaxMappingExtent := FMaxDrawingExtent;
    FEmpty := False;
    Scale; { calculate height and width, kept for compatibility with TGraphic }
    Changed(Self);
  finally
    FreeMem(AFile.Buffer, BufferSize);
  end;
end;

procedure TDxf.LoadFromStreamCompact(Stream: TStream);
var
  i, Count, Version: Smallint;
  ID: Longint;
  NewEntity: PIntDxfEntity;
  NewLineType: PIntDxfLineType;
  NewLayer: PIntDxfLayer;

  // older versions used Extended which was completely unneccessary
  procedure ReadSingle(var Sng: Single);
  var
    Ext: Extended;
  begin
    if Version < 1 then begin
      Stream.Read(Ext, SizeOf(Extended));
      Sng := Ext
    end
    else
      Stream.Read(Sng, SizeOf(Sng))
  end;

  procedure ReadTXYZPoint(var xyz: TXYZPoint);
  begin
    ReadSingle(xyz.x);
    ReadSingle(xyz.y);
    ReadSingle(xyz.z);
  end;

  procedure ReadEntity(var ent: TIntDxfEntity);
  var
    k, PosInc: Integer;
  begin
    Stream.Read(ent.Layer, SizeOf(SmallInt));
    Stream.Read(ent.V.Kind, SizeOf(TDxfEntityKind));
		case ent.V.Kind of
    dxfCircle: begin
        ReadTXYZPoint(ent.V.cCenter);
        ReadSingle(ent.V.cRadius);
        if Version < 2 then begin
          // used to advance stream pointer to end of structure
          if Version < 1 then
            PosInc := SizeOf(Extended) * 2
          else
            PosInc := SizeOf(Single) * 2;
          Stream.Position := Stream.Position + PosInc;
        end;
      end;
    dxfArc: begin
        ReadTXYZPoint(ent.V.aCenter);
        ReadSingle(ent.V.aRadius);
        ReadSingle(ent.V.BegAngle);
        ReadSingle(ent.V.EndAngle);
      end;
    dxfLine: begin
        ReadTXYZPoint(ent.V.Point1);
        ReadTXYZPoint(ent.V.Point2);
      end;
    dxfLWPolyline: begin
        Stream.Read(ent.V.lpVertexCount, SizeOf(SmallInt));
        Stream.Read(ent.V.lpClosed, SizeOf(Boolean));
        ent.V.lpVertices := nil;
        LWPolylineChangeVertexCount(@ent, ent.V.lpVertexCount);
        for k := 0 to Pred(ent.V.lpVertexCount) do
          Stream.Read(ent.V.lpVertices^[k], SizeOf(TIntDxfVertex));
      end;
  	end;
    if Version >= 2 then
      Stream.Read(ent.LineType, SizeOf(SmallInt))
    else
      ent.LineType := LayerLineType;
  end;
begin
  ClearLists;
  Stream.Seek(0, 0);
  Stream.Read(ID, SizeOf(Longint));
  if ID <> DXFFileID then
    raise Exception.Create('This is not a compact DXF');
	Stream.Read(Version, SizeOf(Smallint));
	if Version > DXFFileVer then
		raise Exception.Create('Invalid DXF version');

	{ begin reading data }
	ReadSingle(FAngleBase);
	Stream.Read(FAngleDirection, SizeOf(FAngleDirection));
	Stream.Read(Count, SizeOf(Smallint));
	for i := 1 to Count do begin
		New(NewEntity);
		ReadEntity(NewEntity^);
		FDrawList.Add(NewEntity);
	end;
	Stream.Read(Count, SizeOf(Smallint));
	for i := 1 to Count do begin
		New(NewLineType);
		Stream.Read(NewLineType^, SizeOf(TIntDxfLineType));
		FLineTypeList.Add(NewLineType);
	end;
	Stream.Read(Count, SizeOf(Smallint));
	for i := 1 to Count do begin
		New(NewLayer);
		Stream.Read(NewLayer^, SizeOf(TIntDxfLayer));
		FLayerList.Add(NewLayer);
	end;
	ReadTXYZPoint(FMaxDrawingExtent);
	ReadTXYZPoint(FMaxMappingExtent);
	ReadTXYZPoint(FMinDrawingExtent);
	ReadTXYZPoint(FMinMappingExtent);
	ReadSingle(FScaleFactor);

	FEmpty := False;
	Scale; { calculate height and width, kept for compatibility with TGraphic }
	Changed(Self);
end;

procedure TDxf.LWPolylineChangeVertexCount(const Entity: Pointer;
  const VCount: Integer);
begin
  with PIntDxfEntity(Entity)^ do begin
    V.lpVertexCount := VCount;
    ReallocMem(V.lpVertices, SizeOf(TIntDxfVertex) * V.lpVertexCount);
    FillChar(V.lpVertices^, SizeOf(TIntDxfVertex) * V.lpVertexCount, #0);
  end;
end;

procedure TDxf.MovePoint(var P: TXYZPoint; const Data: Pointer);
var
  AdjustBy: PXYZPoint;
begin
  AdjustBy := PXYZPoint(Data);
  P.x := P.x + AdjustBy^.x;
  P.y := P.y + AdjustBy^.y;
  P.z := P.z + AdjustBy^.z;
end;

{$IFDEF DXFPRINT}
procedure TDxf.Print(Report: TBaseReport; Sideways: Boolean; X, Y, Scale: Single);
const
  PrintLineWidthDxf = -10075;
var
	bx, by: Single;
	BAngle, EAngle: Single;
	i, j, Layer: Integer;
  PenStyle: TPenStyle;

  function GetX(const p: TXYZPoint): Double;
  begin
    if Sideways then
      Result := p.y * Scale + by
    else
      Result := p.x * Scale + bx;
  end;

  function GetY(const p: TXYZPoint): Double;
  begin
    if Sideways then
      Result := p.x * Scale + bx
    else
      Result := -p.y * Scale + by;
  end;

  procedure DrawLineWithThickness(const p1, p2: TXYZPoint;
    const Width1, Width2: Single);
  var
    m: Integer;
    SidePoints: array[0..3] of TXYZPoint;
    CanvasSidePoints: array[0..3] of TPoint;
  begin
    CalcLineWithThicknessPolygon(p1, p2, Width1, Width2, SidePoints);

    for m := Low(SidePoints) to High(SidePoints) do
      CanvasSidePoints[m] := Report.CreatePoint(GetX(SidePoints[m]), GetY(SidePoints[m]));

    Report.SetBrush(clBlack, bsSolid, nil);

    Report.Polygon(CanvasSidePoints);

    Report.SetBrush(clWhite, bsClear, nil);
  end;

  procedure DrawLine(const Point1, Point2: TXYZPoint;
    const StartWidth, EndWidth: Single);
  begin
    if (StartWidth <> 0) or (EndWidth <> 0) then
      DrawLineWithThickness(Point1, Point2, StartWidth, EndWidth)
    else if PenStyle <> psSolid then
      PrintDottedLine(Report, GetX(Point1), GetY(Point1), GetX(Point2), GetY(Point2),
          DashLength)
    else begin
      Report.MoveTo(GetX(Point1), GetY(Point1));
      Report.LineTo(GetX(Point2), GetY(Point2));
    end;
  end;
begin
	if Empty then Exit;
  PenStyle := psSolid; // to remove compiler warning

	with Report do begin
		if Sideways then begin
			bx := Y - FMinMappingExtent.x * Scale;
			by := X - FMinMappingExtent.y * Scale;
		end
		else begin
			bx := X - FMinMappingExtent.x * Scale;
			by := Y + (FMaxMappingExtent.y - FMinMappingExtent.y) * Scale; { reverse Y axis }
		end;
		Layer := -1;

		SetBrush(clWhite, bsClear, nil);
		SetPen(clBlack, psSolid, PrintLineWidthDxf, pmCopy);

		for i := 0 to FDrawList.Count - 1 do begin

			if Layer <> DrawList[i]^.Layer then begin
				Layer := DrawList[i]^.Layer;
			end;
  		PenStyle := GetPenStyleByLayerAndLineType(Layer, DrawList[i]^.LineType);

			with DrawList[i]^ do begin
				if V.Kind = dxfCircle then begin
					if Sideways then begin
            if PenStyle <> psSolid then
              PrintDottedCircle(Report,
	  						V.cCenter.y * Scale + by,
		  					V.cCenter.x * Scale + bx,
			  				V.cRadius * Scale,
				  			DashLength)
            else
  						Ellipse(
	  						(V.cCenter.y - V.cRadius) * Scale + by,
		  					(V.cCenter.x - V.cRadius) * Scale + bx,
			  				(V.cCenter.y + V.cRadius) * Scale + by,
				  			(V.cCenter.x + V.cRadius) * Scale + bx)
          end
					else begin
            if PenStyle <> psSolid then
              PrintDottedCircle(Report,
	  						V.cCenter.x * Scale + bx,
		  					-V.cCenter.y * Scale + by,
			  				V.cRadius * Scale,
				  			DashLength)
            else
  						Ellipse(
	  						(V.cCenter.x - V.cRadius) * Scale + bx,
		  					-(V.cCenter.y - V.cRadius) * Scale + by,
			  				(V.cCenter.x + V.cRadius) * Scale + bx,
				  			-(V.cCenter.y + V.cRadius) * Scale + by);
          end;
				end
				else if V.Kind = dxfArc then begin
          if FAngleDirection = 1 then begin
            BAngle := -V.BegAngle;
            EAngle := -V.EndAngle;
          end
          else begin
            BAngle := V.BegAngle;
            EAngle := V.EndAngle;
          end;
          BAngle := (BAngle + FAngleBase) / 180 * Pi;
          EAngle := (EAngle + FAngleBase) / 180 * Pi;
          if PenStyle <> psSolid then begin
            if Sideways then begin
              PrintDottedCirclularArc(Report,
	  						V.cCenter.y * Scale + by,
		  					V.cCenter.x * Scale + bx,
			  				V.cRadius * Scale,
                BAngle + Pi / 2, EAngle + Pi / 2,
				  			DashLength)
            end
            else begin
              PrintDottedCirclularArc(Report,
     		  				V.cCenter.x * Scale + bx,
		    					-V.cCenter.y * Scale + by,
		  	  				V.cRadius * Scale,
                  BAngle, EAngle,
  				  			DashLength)
            end;
          end
          else begin
            if Sideways then
              Arc(
                (V.aCenter.y - V.aRadius) * Scale + by,
                (V.aCenter.x - V.aRadius) * Scale + bx,
                (V.aCenter.y + V.aRadius) * Scale + by,
                (V.aCenter.x + V.aRadius) * Scale + bx,
                (V.aCenter.y + V.aRadius * sin(BAngle)) * Scale + by,
                (V.aCenter.x + V.aRadius * cos(BAngle)) * Scale + bx,
                (V.aCenter.y + V.aRadius * sin(EAngle)) * Scale + by,
                (V.aCenter.x + V.aRadius * cos(EAngle)) * Scale + bx
                )
            else
              Arc(
                (V.aCenter.x - V.aRadius) * Scale + bx,
                -(V.aCenter.y - V.aRadius) * Scale + by,
                (V.aCenter.x + V.aRadius) * Scale + bx,
                -(V.aCenter.y + V.aRadius) * Scale + by,
                (V.aCenter.x + V.aRadius * cos(BAngle)) * Scale + bx,
                -(V.aCenter.y + V.aRadius * sin(BAngle)) * Scale + by,
                (V.aCenter.x + V.aRadius * cos(EAngle)) * Scale + bx,
                -(V.aCenter.y + V.aRadius * sin(EAngle)) * Scale + by
                );
          end;
				end
				else if V.Kind = dxfLine then
					DrawLine(V.Point1, V.Point2, 0, 0)
        else if V.Kind = dxfLWPolyline then begin
          for j := 0 to V.lpVertexCount - 2 do
            DrawLine(V.lpVertices^[j].P, V.lpVertices^[j+1].P,
                V.lpVertices^[j].StartWidth, V.lpVertices^[j+1].EndWidth);
          if V.lpClosed and (V.lpVertexCount > 2) then
            DrawLine(V.lpVertices^[V.lpVertexCount - 1].P, V.lpVertices^[0].P,
                V.lpVertices^[V.lpVertexCount - 1].StartWidth, V.lpVertices^[0].EndWidth);
        end;
			end;
		end;
	end; {with}
end;
{$ENDIF}

{ all this to read in a line of text? probably could be cleaned up some }

function TDxf.ReadLine(var AFile: TIntDxfFile): string;
const
  CRLF = [#13, #10];
var
  c: Char;
begin
  with AFile do begin
    Result := '';
    if Offset >= BufferSize then begin
      if (Size - Position) < BufferSize then
        BufferLen := Size - Position
      else
        BufferLen := BufferSize;
			Stream.Read(Buffer^, BufferLen);
      Position := Position + BufferLen;
      Offset := 0;
    end;
    while (Position + Offset - BufferLen) < Size do begin
      c := Buffer[Offset];
      Inc(Offset);
      if Offset >= BufferSize then begin
        if (Size - Position) < BufferSize then
          BufferLen := Size - Position
        else
          BufferLen := BufferSize;
        Stream.Read(Buffer^, BufferLen);
        Position := Position + BufferLen;
        Offset := 0;
      end;
      if c = #13 then begin
        Inc(LineNo);
        Exit;
			end;
      if c = #10 then Continue;
      Result := Result + c;
    end;
    Eof := True;
  end;
end;

procedure TDxf.ResetOrigin;
var
  MinMax: TIntMinMaxPoints;
begin
  GetExtents(MinMax);
  TransformBy(XYZPoint(-MinMax.Max.x, -MinMax.Min.y, 0));
end;

procedure TDxf.SaveToClipboardFormat(var AFormat: Word; var AData: THandle;
		var APalette: HPALETTE);
begin
	raise Exception.Create('TDxf does not yet support saving to clipboard format');
end;

procedure TDxf.SaveToStream(Stream: TStream);
begin
	raise Exception.Create('TDxf does not yet support saving to a stream');
end;

procedure TDxf.SaveToStreamCompact(Stream: TStream);
var
  i: Integer;
  ID: Longint;

  procedure WriteEntity(const ent: TIntDxfEntity);
  var
    k: Integer;

    procedure WriteTXYZPoint(const p: TXYZPoint);
    begin
      Stream.Write(p, SizeOf(TXYZPoint))
    end;

    procedure WriteSingle(const p: Single);
    begin
      Stream.Write(p, SizeOf(Single))
    end;
  begin
    Stream.Write(ent.Layer, SizeOf(SmallInt));
    Stream.Write(ent.V.Kind, SizeOf(TDxfEntityKind));
		case ent.V.Kind of
    dxfCircle: begin
        WriteTXYZPoint(ent.V.cCenter);
        WriteSingle(ent.V.cRadius);
      end;
    dxfArc: begin
        WriteTXYZPoint(ent.V.aCenter);
        WriteSingle(ent.V.aRadius);
        WriteSingle(ent.V.BegAngle);
        WriteSingle(ent.V.EndAngle);
      end;
    dxfLine: begin
        WriteTXYZPoint(ent.V.Point1);
        WriteTXYZPoint(ent.V.Point2);
      end;
    dxfLWPolyline: begin
        Stream.Write(ent.V.lpVertexCount, SizeOf(SmallInt));
        Stream.Write(ent.V.lpClosed, SizeOf(Boolean));
        for k := 0 to Pred(ent.V.lpVertexCount) do
          Stream.Write(ent.V.lpVertices^[k], SizeOf(TIntDxfVertex));
      end;
    end;
    Stream.Write(ent.LineType, SizeOf(SmallInt));
  end;
begin
  Stream.Seek(0, 0);
  ID := DXFFileID;
  Stream.Write(ID, SizeOf(Longint));
  i := DXFFileVer;
	Stream.Write(i, SizeOf(Smallint));

	{ begin writing data }
	Stream.Write(FAngleBase, SizeOf(FAngleBase));
	Stream.Write(FAngleDirection, SizeOf(FAngleDirection));
	i := FDrawList.Count;
	Stream.Write(i, SizeOf(Smallint));
	for i := 0 to FDrawList.Count - 1 do
    WriteEntity(DrawList[i]^);
	i := FLineTypeList.Count;
	Stream.Write(i, SizeOf(Smallint));
	for i := 0 to FLineTypeList.Count - 1 do
		Stream.Write(LineTypeList[i]^, SizeOf(TIntDxfLineType));
	i := FLayerList.Count;
	Stream.Write(i, SizeOf(Smallint));
	for i := 0 to FLayerList.Count - 1 do
		Stream.Write(LayerList[i]^, SizeOf(TIntDxfLayer));
	Stream.Write(FMaxDrawingExtent, SizeOf(TXYZPoint));
	Stream.Write(FMaxMappingExtent, SizeOf(TXYZPoint));
	Stream.Write(FMinDrawingExtent, SizeOf(TXYZPoint));
	Stream.Write(FMinMappingExtent, SizeOf(TXYZPoint));
	Stream.Write(FScaleFactor, SizeOf(FScaleFactor));
	Stream.Size := Stream.Position;
end;

{ calculate height & width based on ScaleFactor }

procedure TDxf.Scale;
var
  w, h: Longint;
begin
  h := Round(Abs(FScaleFactor * (FMaxMappingExtent.y - FMinMappingExtent.y)));
  w := Round(Abs(FScaleFactor * (FMaxMappingExtent.x - FMinMappingExtent.x)));
  if h > 32767 then
    FHeight := 32767
  else
    FHeight := h;
  if w > 32767 then
    FWidth := 32767
  else
    FWidth := w;
end;

{ Set the height = set the ScaleFactor }

procedure TDxf.SetHeight(Value: Integer);
begin
  FScaleFactor := Value / Abs(FMaxMappingExtent.y - FMinMappingExtent.y);
  Scale;
  Changed(Self);
end;

{ set the mapping extent and recalc height & width }

procedure TDxf.SetMaxMappingExtent(Value: TXYZPoint);
begin
  FMaxMappingExtent := Value;
  Scale;
  Changed(Self);
end;

{ set the mapping extent and recalc height & width }

procedure TDxf.SetMinMappingExtent(Value: TXYZPoint);
begin
  FMinMappingExtent := Value;
  Scale;
  Changed(Self);
end;

{ Set the ScaleFactor and recalc. the height & width }

procedure TDxf.SetScaleFactor(Value: Single);
begin
  FScaleFactor := Value;
  Scale;
  Changed(Self);
end;

{ Set the width = set the ScaleFactor }

procedure TDxf.SetWidth(Value: Integer);
begin
  FScaleFactor := Value / Abs(FMaxMappingExtent.x - FMinMappingExtent.x);
  Scale;
  Changed(Self);
end;

procedure TDxf.ZoomToExtents;
var
  MinMax: TIntMinMaxPoints;
begin
  GetExtents(MinMax);

  FMinDrawingExtent := MinMax.Min;
  FMaxDrawingExtent := MinMax.Max;
  FMinMappingExtent := MinMax.Min;
  FMaxMappingExtent := MinMax.Max;
end;

function TDxf.GetMappingDepth: Single;
begin
  Result := MaxMappingExtent.z - MinMappingExtent.z;
end;

function TDxf.GetMappingHeight: Single;
begin
  Result := MaxMappingExtent.y - MinMappingExtent.y;
end;

function TDxf.GetMappingWidth: Single;
begin
  Result := MaxMappingExtent.x - MinMappingExtent.x;
end;

procedure TDxf.ScaleBy(Factor: TXYZPoint);
var
  i: Integer;
begin
  for i := 0 to Pred(FDrawList.Count) do
    EnumerateEntityPoints(DrawList[i], @Factor, ScalePoint);

  ScalePoint(FMaxDrawingExtent, @Factor);
  ScalePoint(FMaxMappingExtent, @Factor);
  ScalePoint(FMinDrawingExtent, @Factor);
	ScalePoint(FMinMappingExtent, @Factor);
end;

procedure TDxf.DeleteEntitiesInRange(FromPoint, ToPoint: TXYZPoint);
var
  i, j: Integer;
  DeleteEntity: Boolean;

  function PointContained(Point: TXYZPoint): Boolean;
  begin
    Result := (Point.x >= FromPoint.x) and (Point.y >= FromPoint.y)
        and (Point.z >= FromPoint.z) and (Point.x <= ToPoint.x)
        and (Point.y <= ToPoint.y) and (Point.z <= ToPoint.z);
  end;
begin
  for i := FDrawList.Count - 1 downto 0 do begin
    with DrawList[i]^ do
      case V.Kind of
      dxfCircle: DeleteEntity := PointContained(V.cCenter);
      dxfArc: DeleteEntity := PointContained(V.aCenter);
      dxfLine: DeleteEntity := PointContained(V.Point1) and PointContained(V.Point2);
      dxfLWPolyline: begin
          DeleteEntity := True;
          for j := 0 to Pred(V.lpVertexCount) do
            DeleteEntity := DeleteEntity and PointContained(V.lpVertices^[j].P);
        end;
      else DeleteEntity := False;
      end;
    if DeleteEntity then begin
      FreeEntity(FDrawList[i]);
      FDrawList.Delete(i);
    end;
  end;
end;

procedure TDxf.FreeEntity(PEntity: PIntDxfEntity);
begin
  with PEntity^ do
    if (V.Kind = dxfLWPolyline) and (V.lpVertices <> nil) then
      FreeMem(V.lpVertices);
  Dispose(PEntity);
end;

procedure TDxf.TransformBy(Delta: TXYZPoint);
var
  i: Integer;
begin
  for i := 0 to Pred(FDrawList.Count) do
    EnumerateEntityPoints(DrawList[i], @Delta, MovePoint);

  MovePoint(FMinDrawingExtent, @Delta);
  MovePoint(FMaxDrawingExtent, @Delta);

  MovePoint(FMinMappingExtent, @Delta);
  MovePoint(FMaxMappingExtent, @Delta);
end;

procedure TDxf.ScalePoint(var P: TXYZPoint; const Data: Pointer);
var
  Factor: PXYZPoint;
begin
  Factor := PXYZPoint(Data);
  P.x := P.x * Factor^.x;
  P.y := P.y * Factor^.y;
  P.z := P.z * Factor^.z;
end;

initialization
// WARNING THIS CODE CAUSES CRASH ON DELPHI IDE EXIT
// (ALSO I DON'T THINK IT WORKS ANYWAY)
//	TPicture.RegisterFileFormat('dxf', 'Dxf files', TDxf);

end.

