{
DESCRIPTION:
This unit tests QSort.pas.

COPYRIGHT:
Copyright 2002 Rio Software, LLC.  All rights reserved.

HISTORY:
Created on 8/19/2002 by Colin Patrick Sarsfield & Andrew Cray Vogan
}

unit VQSort;

interface

uses
  Classes, TestFrameWork;

type
  TTestQSort = class(TTestCase)
  private
    FCompareCount: Integer;
    FExchangeCount: Integer;
    FValues: array of Integer;

    function CompareItems(const a, b: Integer): Integer;
    procedure ExchangeItems(const a, b: Integer);
    function IsSorted(TestArray: array of Integer): Boolean;
    procedure RunQuickSort(CutOff: Integer = 10);
  published
    procedure TestIdentical;
    procedure TestIsSorted;
    procedure TestRandom;
    procedure TestSorted;
    procedure TestSortedWithoutIdenticals;
    procedure TestWorstCase;
  end;

implementation

uses
  SysUtils, Math, QSort;

const
  TestCount = 159;

var
  AverageCompareAllowance: Integer;

{ TTestQSort }

function TTestQSort.CompareItems(const a, b: Integer): Integer;
begin
  CompareItems := Ord(FValues[a] > FValues[b]) - Ord(FValues[b] > FValues[a]);
  Inc(FCompareCount);
end;

procedure TTestQSort.ExchangeItems(const a, b: Integer);
var
  c: Integer;
begin
  c := FValues[a];
  FValues[a] := FValues[b];
  FValues[b] := c;

  Inc(FExchangeCount);
end;

function TTestQSort.IsSorted(TestArray: array of Integer): Boolean;
var
  i: Integer;
begin
  IsSorted := False;
  for i := Succ(Low(TestArray)) to High(TestArray) do
    if TestArray[Pred(i)] > TestArray[i] then
      Exit;
  IsSorted := True;
end;

procedure TTestQSort.RunQuickSort(CutOff: Integer);
begin
  FCompareCount := 0;
  FExchangeCount := 0;
  QuickSort(Low(FValues), High(FValues), CompareItems, ExchangeItems, CutOff);
end;

procedure TTestQSort.TestIdentical;
var
  i: Integer;
begin
  SetLength(FValues, TestCount);
  for i := 0 to High(FValues) do
    FValues[i] := 199482;
  RunQuickSort;
  Check(FCompareCount < AverageCompareAllowance,
      Format('TestIdentical: %d compares; %d exchanges', [FCompareCount, FExchangeCount]));
end;

procedure TTestQSort.TestIsSorted;
begin
  Check(IsSorted([1, 2, 3, 4, 5]), 'A sorted array');
  Check(not IsSorted([3, 2, 1, 4, 5]), 'Unsorted');
  Check(not IsSorted([1, 2, 3, 4, 2]), 'last out of order');
  Check(not IsSorted([3, 2, 3, 4, 5]), 'first out of order');
  Check(IsSorted([1, 1, 1, 1, 1]), 'all same value');
end;

procedure TTestQSort.TestRandom;
var
  i: Integer;
begin
  SetLength(FValues, TestCount);
  for i := 0 to High(FValues) do
    FValues[i] := RandomRange(Low(Integer), High(Integer));
  RunQuickSort;
  Check(IsSorted(FValues));
  Check(FCompareCount < AverageCompareAllowance,
      Format('TestRandom: %d compares; %d exchanges', [FCompareCount, FExchangeCount]));
end;

procedure TTestQSort.TestSorted;
var
  i: Integer;
begin
  SetLength(FValues, TestCount);
  FValues[0] := 1;
  for i := 1 to High(FValues) do
    FValues[i] := FValues[Pred(i)] + RandomRange(0, 2);
  RunQuickSort;
  Check(IsSorted(FValues));
  Check(FCompareCount < AverageCompareAllowance,
      Format('TestSorted (maybe some identicals): %d compares; %d exchanges', [FCompareCount, FExchangeCount]));
end;

procedure TTestQSort.TestSortedWithoutIdenticals;
var
  i: Integer;
begin
  SetLength(FValues, TestCount);
  for i := 0 to High(FValues) do
    FValues[i] := i * 2;
  RunQuickSort;
  Check(IsSorted(FValues));
  Check(FCompareCount < AverageCompareAllowance,
      Format('TestSorted (no identicals): %d compares; %d exchanges', [FCompareCount, FExchangeCount]));
end;

procedure TTestQSort.TestWorstCase;
const
  Size = 9;
var
  i, DrawIdx, DrawValue, SizeFact,
  MaxCompares, MaxExchanges, MaxPerm: Integer;

  function DumpArray: String;
  var
    m: Integer;
  begin
    Result := '';
    for m := Low(FValues) to High(FValues) do
      Result := Result + IntToStr(FValues[m]) + ' ';
  end;

  procedure InitByI;
  var
    j, k, Perm: Integer;
  begin
    Perm := i;
    for j := 0 to Pred(Size) do
      FValues[j] := Succ(j);
    for j := 0 to Pred(Size) do begin
      DrawIdx := j + Perm mod (Size - j);
      DrawValue := FValues[DrawIdx];
      for k := Pred(DrawIdx) downto j do
        FValues[Succ(k)] := FValues[k];
      FValues[j] := DrawValue;
      Perm := Perm div (Size - j);
    end;
  end;
begin
  MaxPerm := 0; // to make compiler happy 
  MaxCompares := 0;
  MaxExchanges := 0;
  SetLength(FValues, Size);

  SizeFact := 1;
  for i := 2 to Size do
    SizeFact := SizeFact * i;

  for i := 0 to Pred(SizeFact) do begin
    InitByI;
    RunQuickSort(0);
    if (FCompareCount > MaxCompares)
        or ((FCompareCount = MaxCompares) and (FExchangeCount >= MaxExchanges)) then begin
      MaxCompares := FCompareCount;
      MaxExchanges := FExchangeCount;
      MaxPerm := i;
    end;
  end;
  i := MaxPerm;
  InitByI;
  Check(DumpArray = '2 6 4 8 1 3 5 7 9 ', 'Worst case sorting length 9 array');
  Check(MaxCompares < Sqr(Size), Format('MaxCompares = %d, MaxExchanges = %d',
      [MaxCompares, MaxExchanges]));
end;

initialization
  AverageCompareAllowance := Round(2 * TestCount * Ln(TestCount));

  TestFramework.RegisterTest(TTestQSort.Suite);
end.

